from pyranges.pyranges import PyRanges

from tests.helpers import assert_df_equal
import pytest

from io import StringIO

import pandas as pd


@pytest.fixture
def background():
    c = """Chromosome Start End Name Score Strand
chr1	226987603	226987628	U0	0	-
chr8	38747236	38747261	U0	0	+
chr15	26105493	26105518	U0	0	+"""


    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    print(df)
    return PyRanges(df)

@pytest.fixture
def chip():

    c = """Chromosome Start End Name Score Strand
chr1	226987592	226987617	U0	0	+
chr8	38747226	38747251	U0	0	-
chr15	26105515	26105540	U0	0	+"""


    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    # print(df)
    return PyRanges(df)



@pytest.fixture
def expected_result_regular_intersection():

    c = """chr1	226987603	226987617	U0	0	+
chr8	38747236	38747251	U0	0	-
chr15	26105515	26105518	U0	0	+"""

    return PyRanges(pd.read_table(StringIO(c), header=None, names="Chromosome Start End Name Score Strand".split()))


def test_intersect_granges(chip, background, expected_result_regular_intersection):

    print("chip " * 3)
    print(chip)

    print("background " * 3)
    print(background)

    print("expected_result " * 3)
    print(expected_result_regular_intersection)

    result = chip.intersection(background, strandedness=False)

    print(result)

    assert assert_df_equal(result.df, expected_result_regular_intersection.df)





import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture
def expected_result_unstranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 1 2 f 0 + 1
1 chr1 5 7 h 0 - 1 2 f 0 + 3
2 chr1 8 9 h 0 + 6 7 f 0 - 1"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_unstranded(f1, f2, expected_result_unstranded):

    result = f1.nearest(f2, strandedness=False, suffix="_b", how="previous_nonoverlapping")

    print(result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_unstranded.df)


@pytest.fixture
def expected_result_opposite_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 10246 10445 HWI-ST216_313:3:1207:4315:142177 1 + 9988 10187 HWI-ST216:427:D29R1ACXX:2:1205:6095:16532 1 - 59
1 chr1 110246 110445 HWI-ST216_313:3:1207:4315:142177 1 + 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 90089"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_opposite_stranded(chip_10_plus_one, input_10, expected_result_opposite_stranded):

    result = chip_10_plus_one.nearest(input_10, strandedness="opposite", suffix="_b", how="previous_nonoverlapping")

    print("result", result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_opposite_stranded.df)


@pytest.fixture
def expected_result_same_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 9988 10187 HWI-ST216:427:D29R1ACXX:2:1205:6095:16532 1 - 54
1 chr1 110246 110445 HWI-ST216_313:3:1207:4315:142177 1 + 16109 16308 HWI-ST216:427:D29R1ACXX:2:2110:12286:25379 1 + 93938"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_same_stranded(chip_10_plus_one, input_10, expected_result_same_stranded):

    result = chip_10_plus_one.nearest(input_10, strandedness="same", suffix="_b", how="previous_nonoverlapping")

    print("result", result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_same_stranded.df)



from io import StringIO

from tests.helpers import assert_df_equal
from pyranges.pyranges import PyRanges

import pandas as pd

import pytest

@pytest.fixture
def expected_result_unstranded(names):

    c = """chr1    9916    9988    HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    9988    HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    9988    HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    9988    HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    9988    HWI-ST216_313:3:1204:5599:113305        1       -"""

    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)


@pytest.fixture
def expected_result_self_unstranded(names):

    c = """chr1    9916    9988    HWI-ST216_313:3:1203:10227:6568 1
chr1    9939    9988    HWI-ST216_313:3:2301:15791:16298        1
chr1    9951    9988    HWI-ST216_313:3:2205:20086:33508        1
chr1    9953    9988    HWI-ST216_313:3:1305:6975:102491        1
chr1    9978    9988    HWI-ST216_313:3:1204:5599:113305        1"""

    df = pd.read_table(StringIO(c), sep="\s+", names=names[:-1], header=None)
    return PyRanges(df)


@pytest.fixture
def expected_result_no_strand_plus_one(names):

    c = """chr1    9916    9988    HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    9988    HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    9988    HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    9988    HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    9988    HWI-ST216_313:3:1204:5599:113305        1       -
chr1    10348   10445   HWI-ST216_313:3:1207:4315:142177        1       +"""

    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)



@pytest.fixture
def expected_result_same_strand(names):

    c = """chr1    9916    9988    HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    10073   HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    9988    HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    10073   HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    9988    HWI-ST216_313:3:1204:5599:113305        1       -
chr1    10024   10073   HWI-ST216_313:3:2201:5209:155139        1       +
chr1    10348   10440   HWI-ST216_313:3:1302:4516:156396        1       -
chr1    10272   10280   HWI-ST216_313:3:1207:4315:142177        1       +"""

    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)

@pytest.fixture
def expected_result_same_strand_plus_one(names):

    c = """chr1    9916    9988    HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    10073   HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    9988    HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    10073   HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    9988    HWI-ST216_313:3:1204:5599:113305        1       -
chr1    10024   10073   HWI-ST216_313:3:2201:5209:155139        1       +
chr1    10348   10440   HWI-ST216_313:3:1302:4516:156396        1       -
chr1    10272   10280   HWI-ST216_313:3:1207:4315:142177        1       +
chr1    10348   10445   HWI-ST216_313:3:1207:4315:142177        1       +"""

    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)

@pytest.fixture
def expected_result_opposite_strand(names):

    c = """chr1    9916    10073   HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    9988    HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    10073   HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    9988    HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    10073   HWI-ST216_313:3:1204:5599:113305        1       -
chr1    10001   10073   HWI-ST216_313:3:1102:14019:151362       1       -
chr1    10272   10280   HWI-ST216_313:3:2207:7406:122346        1       -
chr1    10272   10280   HWI-ST216_313:3:1302:4516:156396        1       -
chr1    10348   10445   HWI-ST216_313:3:1207:4315:142177        1       +"""


    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)

@pytest.fixture
def expected_result_opposite_strand_plus_one(names):

    c = """chr1    9916    10073   HWI-ST216_313:3:1203:10227:6568 1       -
chr1    9939    9988    HWI-ST216_313:3:2301:15791:16298        1       +
chr1    9951    10073   HWI-ST216_313:3:2205:20086:33508        1       -
chr1    9953    9988    HWI-ST216_313:3:1305:6975:102491        1       +
chr1    9978    10073   HWI-ST216_313:3:1204:5599:113305        1       -
chr1    10001   10073   HWI-ST216_313:3:1102:14019:151362       1       -
chr1    10272   10280   HWI-ST216_313:3:2207:7406:122346        1       -
chr1    10272   10280   HWI-ST216_313:3:1302:4516:156396        1       -
chr1    10348   10445   HWI-ST216_313:3:1207:4315:142177        1       +
chr1    110246  110445  HWI-ST216_313:3:1207:4315:142177        1       +"""


    df = pd.read_table(StringIO(c), sep="\s+", names=names, header=None)
    return PyRanges(df)



def test_advanced_subtraction_unstranded(chip_10, input_10, expected_result_unstranded):

    result = chip_10.subtraction(input_10)

    print(result)

    assert_df_equal(result.df, expected_result_unstranded.df)


def test_advanced_subtraction_same_strand(chip_10, input_10, expected_result_same_strand):

    result = chip_10.subtraction(input_10, strandedness="same")

    assert_df_equal(result.df, expected_result_same_strand.df)


def test_advanced_subtraction_opposite_strand(chip_10, input_10, expected_result_opposite_strand):

    result = chip_10.subtraction(input_10, strandedness="opposite")

    assert_df_equal(result.df, expected_result_opposite_strand.df)

def test_advanced_subtraction_no_strand_plus_one(chip_10_plus_one, input_10, expected_result_no_strand_plus_one):

    result = chip_10_plus_one.subtraction(input_10)

    assert_df_equal(result.df, expected_result_no_strand_plus_one.df)

def test_advanced_subtraction_same_strand_plus_one(chip_10_plus_one, input_10, expected_result_same_strand_plus_one):

    result = chip_10_plus_one.subtraction(input_10, strandedness="same")

    assert_df_equal(result.df, expected_result_same_strand_plus_one.df)

def test_advanced_subtraction_opposite_strand_plus_one(chip_10_plus_one, input_10, expected_result_opposite_strand_plus_one):

    result = chip_10_plus_one.subtraction(input_10, strandedness="opposite")

    assert_df_equal(result.df, expected_result_opposite_strand_plus_one.df)


def test_advanced_subtraction_unstranded_self_no_strand(chip_10_no_strand, input_10, expected_result_self_unstranded):

    result = chip_10_no_strand.subtraction(input_10)

    assert_df_equal(result.df, expected_result_self_unstranded.df)


def test_advanced_subtraction_unstranded_other_no_strand(chip_10, input_10_no_strand, expected_result_unstranded):

    result = chip_10.subtraction(input_10_no_strand)

    assert_df_equal(result.df, expected_result_unstranded.df)




import pytest

from pyranges.pyranges import PyRanges

import pandas as pd

from io import StringIO

def test_getsetattr(chip_10):

    chip_10.whatevz = range(len(chip_10))

    assert (chip_10.whatevz == pd.Series(range(len(chip_10)))).all()



def test_getsetattr_fails(chip_10):

    with pytest.raises(Exception):
        chip_10.whatevz = range(len(chip_10) + 1)


def test_getsetattr_with_str(chip_10):

    chip_10.whatevz = "whatevz"

    assert (chip_10.whatevz == pd.Series(["whatevz"] * len(chip_10))).all()



import pytest

from tests.helpers import assert_df_equal
from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture
def cs():

    cs = pr.load_dataset("chipseq")
    return cs


@pytest.fixture
def bg():

    bg = pr.load_dataset("chipseq_background")
    return bg


@pytest.fixture
def expected_result_regular_overlap():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b
chr8	38747226	38747251	U0	0	-	38747236	38747261	U0	0	+
chr1	226987592	226987617	U0	0	+	226987603	226987628	U0	0	-
chr15	26105515	26105540	U0	0	+	26105493	26105518	U0	0	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+"))


def test_advanced_overlap(cs, bg, expected_result_regular_overlap):

    result = cs.join(bg)

    print("result")
    print(result)
    print("expected_result_regular_overlap")
    print(expected_result_regular_overlap)

    assert assert_df_equal(expected_result_regular_overlap.df, result.df)

@pytest.fixture
def expected_result_regular_overlap_intersection():

    c = """Chromosome Start End Start_a End_a Name_a Score_a Strand Start_b End_b Name_b Score_b Strand_b
chr1 226987603 226987617 226987592 226987617 U0 0 + 226987603 226987628 U0 0 -
chr8 38747236 38747251 38747226 38747251 U0 0 - 38747236 38747261 U0 0 +
chr15 26105515 26105518 26105515 26105540 U0 0 + 26105493 26105518 U0 0 +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+"))

def test_advanced_overlap_new_pos_intersect(cs, bg, expected_result_regular_overlap_intersection):

    result = cs.join(bg, new_pos="intersection")

    assert assert_df_equal(result.df, expected_result_regular_overlap_intersection.df)
    # assert expected_result_regular_overlap_intersection.df.equals(result.df)



@pytest.fixture
def expected_result_overlap_same_strand():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b
chr15	26105515	26105540	U0	0	+	26105493	26105518	U0	0	+"""

    df = pd.read_table(StringIO(c), header=0, sep="\s+")
    return PyRanges(df)

def test_advanced_overlap_stranded(cs, bg, expected_result_overlap_same_strand):

    result = cs.join(bg, strandedness="same")

    print("result")
    print(result)
    print("expected result")
    print(expected_result_overlap_same_strand)

    assert assert_df_equal(expected_result_overlap_same_strand.df, result.df)


@pytest.fixture
def expected_result_overlap_opposite_strand():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b
chr8	38747226	38747251	U0	0	-	38747236	38747261	U0	0	+
chr1	226987592	226987617	U0	0	+	226987603	226987628	U0	0	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+"))


def test_advanced_overlap_opposite_stranded(cs, bg, expected_result_overlap_opposite_strand):

    result = cs.join(bg, strandedness="opposite")

    print(result.df)
    print(expected_result_overlap_opposite_strand)

    assert assert_df_equal(expected_result_overlap_opposite_strand.df, result.df)



@pytest.fixture
def expected_result_overlap_opposite_strand_suffixes():

    c = """Chromosome Start End Name Score Strand Starthiya Endhiya Namehiya Scorehiya Strandhiya
chr8	38747226	38747251	U0	0	-	38747236	38747261	U0	0	+
chr1	226987592	226987617	U0	0	+	226987603	226987628	U0	0	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+"))


def test_advanced_overlap_opposite_stranded_suffixes(cs, bg, expected_result_overlap_opposite_strand_suffixes):

    result = cs.join(bg, strandedness="opposite", suffixes=["", "hiya"])

    print(result.df)
    print(expected_result_overlap_opposite_strand_suffixes)

    assert assert_df_equal(expected_result_overlap_opposite_strand_suffixes.df, result.df)


@pytest.fixture
def expected_result_regular_overlap_union():

    c = """Chromosome Start End Start_a End_a Name_a Score_a Strand Start_b End_b Name_b Score_b Strand_b
0 chr1 226987592 226987628 226987592 226987617 U0 0 + 226987603 226987628 U0 0 -
1 chr8 38747226 38747261 38747226 38747251 U0 0 - 38747236 38747261 U0 0 +
2 chr15 26105493 26105540 26105515 26105540 U0 0 + 26105493 26105518 U0 0 +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+"))

def test_advanced_overlap_new_pos_union(cs, bg, expected_result_regular_overlap_union):

    result = cs.join(bg, new_pos="union")

    print("result")
    print(result.df.to_csv(sep=" "))
    print("expected_result_regular_overlap")
    print(expected_result_regular_overlap_union)

    assert assert_df_equal(expected_result_regular_overlap_union.df, result.df)




import pytest

from pyranges.pyranges import PyRanges

from tests.helpers import assert_df_equal
import pandas as pd

from io import StringIO


@pytest.fixture
def simple_gr1():

    c = """Chromosome Start End Strand Score
chr1 3 6 + 5
chr1 5 7 - 7
chr1 8 9 + 1"""

    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    return PyRanges(df)



@pytest.fixture
def simple_gr2():

    c = """Chromosome Start End Strand Score
chr1 1 2 + 1
chr1 6 7 - 2"""
    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    return PyRanges(df)


@pytest.fixture
def expected_result_intersection_simple_granges():

    c = """Chromosome Start End Strand Score
chr1    6   7   - 7"""

    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    return PyRanges(df)



def test_intersect_simple_granges(simple_gr1, simple_gr2, expected_result_intersection_simple_granges):

    result = simple_gr1.intersection(simple_gr2, strandedness=False)

    print("result\n", result.df)
    print("expected_result_intersection_simple_granges\n",
          expected_result_intersection_simple_granges.df)

    assert assert_df_equal(result.df, expected_result_intersection_simple_granges.df)


@pytest.fixture
def expected_result_same_strand_intersection_simple_granges():

    c = """Chromosome Start End Strand Score
chr1    6   7   - 7"""

    df = pd.read_table(StringIO(c), sep="\s+", header=0)
    return PyRanges(df)


def test_intersect_same_strand_simple_granges(simple_gr1, simple_gr2, expected_result_same_strand_intersection_simple_granges):

    result = simple_gr1.intersection(simple_gr2, strandedness="same")

    print(result)

    assert assert_df_equal(expected_result_same_strand_intersection_simple_granges.df, result.df)



# @pytest.fixture
# def expected_result_opposite_strand_intersection_simple_granges():

#     c = """Chromosome Start End Strand Score
# chr1	3	6	+ 5
# chr1    5   7   - 7
# chr1	8	9	+ 1"""

#     df = pd.read_table(StringIO(c), sep="\s+", header=0)
#     return PyRanges(df)


def test_intersect_opposite_strand_simple_granges(simple_gr1, simple_gr2):

    result = simple_gr1.intersection(simple_gr2, strandedness="opposite")

    print(result)

    assert result.df.empty



import pandas as pd
import numpy as np
from collections import defaultdict

from tabulate import tabulate

from pyranges.methods import (_overlap, _cluster, _tile, _intersection,
                              _coverage, _overlap_write_both,
                              _set_intersection, _set_union, _subtraction,
                              _nearest)

from ncls import NCLS

# try:
#     dummy = profile
# except:
#     profile = lambda x: x


def create_ncls(cdf):

    return NCLS(cdf.Start.values,
                cdf.End.values,
                cdf.index.values)


def create_ncls_dict(df, strand):

    if not strand:
        grpby = df.groupby("Chromosome")
    else:
        grpby = df.groupby("Chromosome Strand".split())

    nclses = {key: create_ncls(cdf) for (key, cdf) in grpby}
    dd = defaultdict(NCLS)
    dd.update(nclses)

    return dd


def create_pyranges_df(seqnames, starts, ends, strands=None):

    if isinstance(seqnames, str):
        seqnames = [seqnames] * len(starts)

    if isinstance(strands, str):
        strands = [strands] * len(starts)

    if strands != None and strands != False:
        columns = [seqnames, starts, ends, strands]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts, ends and strands must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End Strand".split()
    else:
        columns = [seqnames, starts, ends]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts and ends must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End".split()

    idx = range(len(starts))
    series_to_concat = []
    for s in columns:
        s = pd.Series(s, index=idx)
        series_to_concat.append(s)

    df = pd.concat(series_to_concat, axis=1)
    df.columns = colnames

    return df



class PyRanges():

    df = None

    def __init__(self, df=None, seqnames=None, starts=None, ends=None, strands=None):

        if df is False or df is None:
            df = create_pyranges_df(seqnames, starts, ends, strands)
        else:
            assert "Chromosome" in df and "Start" in df and "End" in df
            df.index = range(len(df))

        # using __dict__ to avoid invoking setattr
        if "Strand" not in df:
            self.__dict__["stranded"] = False
        else:
            self.__dict__["stranded"] = True

        self.__dict__["df"] = df

        self.__dict__["__ncls__"] = create_ncls_dict(df, self.stranded)


    def __len__(self):
        return self.df.shape[0]

    def __setattr__(self, column_name, column):

        if column_name in "Chromosome Start End Strand".split():
            raise Exception("The columns Chromosome, Start, End or Strand can not be reset.")
        if column_name == "stranded":
            raise Exception("The stranded attribute is read-only. Create a new PyRanges object instead.")
        if not isinstance(column, str):
            if not len(self.df) == len(column):
                raise Exception("DataFrame and column must be same length.")

            column_to_insert = pd.Series(column, index=self.df.index)
        else:
            column_to_insert = pd.Series(column, index=self.df.index)

        pos = self.df.shape[1]
        if column_name in self.df:
            pos = list(self.df.columns).index(column_name)
            self.df.drop(column_name, inplace=True, axis=1)

        self.df.insert(pos, column_name, column_to_insert)


    def __getattr__(self, name):

        if name in self.df:
            return self.df[name]
        else:
            self.__dict__[name]


    def __getitem__(self, val):

        pd.options.mode.chained_assignment = None
        if isinstance(val, str):
            if val in set(self.df.Chromosome):
                return PyRanges(self.df.loc[self.df.Chromosome == val])
            elif val in "+ -".split():
                return PyRanges(self.df.loc[self.df.Strand == val])
            else:
                raise Exception("Invalid choice for string subsetting PyRanges: {}. Must be either strand or chromosome".format(val))

        elif isinstance(val, tuple):

            # "chr1", 5:10
            if len(val) == 2 and val[0] in self.df.Chromosome.values and isinstance(val[1], slice):
                chromosome, loc = val
                start = loc.start or 0
                stop = loc.stop or max(self.df.loc[self.df.Chromosome == chromosome].End.max(), start)
                idxes = [r[2] for r in self.__ncls__[chromosome, "+"].find_overlap(start, stop)] + \
                        [r[2] for r in self.__ncls__[chromosome, "-"].find_overlap(start, stop)]

                return PyRanges(self.df.loc[idxes])

            # "+", 5:10
            if len(val) == 2 and val[0] in "+ -".split() and isinstance(val[1], slice):
                strand, loc = val
                start = loc.start or 0
                stop = loc.stop or max(self.df.loc[self.df.Chromosome == chromosome].End.max(), start)
                idxes = []
                for chromosome in self.df.Chromosome.drop_duplicates():
                    idxes.extend([r[2] for r in self.__ncls__[chromosome, strand].find_overlap(start, stop)])

                return PyRanges(self.df.loc[idxes])

            # "chr1", "+"
            if len(val) == 2 and val[1] in "+ -".split():

                chromosome, strand = val

                return PyRanges(self.df.loc[(self.df.Chromosome == chromosome) & (self.df.Strand == strand)])

            # "chr1", "+", 5:10
            elif len(val) == 3 and val[0] in self.df.Chromosome.values and val[1] in "+ -".split():

                chromosome, strand, loc = val
                start = loc.start or 0
                stop = loc.stop or max(self.df.loc[self.df.Chromosome == chromosome].End.max(), start)
                idxes = [r[2] for r in self.__ncls__[chromosome, strand].find_overlap(start, stop)]

                return PyRanges(self.df.loc[idxes])

        elif isinstance(val, slice):

            start = val.start or 0
            stop = val.stop or max(self.df.End.max(), start)

            idxes = []
            for it in self.__ncls__.values():
                idxes.extend([r[2] for r in it.find_overlap(start, stop)])

            return PyRanges(self.df.loc[idxes])

        pd.options.mode.chained_assignment = "warn"


    def __str__(self):

        if len(self.df) > 6:
            h = self.df.head(3).astype(object)
            t = self.df.tail(3).astype(object)
            m = self.df.head(1).astype(object)
            m.loc[:,:] = "..."
            m.index = ["..."]
            s = pd.concat([h, m, t])
        else:
            s = self.df

        str_repr = tabulate(s, headers='keys', tablefmt='psql', showindex=False) + \
                                        "\nPyRanges object has {} sequences from {} chromosomes.".format(self.df.shape[0], len(set(self.df.Chromosome)))
        return str_repr


    def __repr__(self):

        return str(self)


    def overlap(self, other, strandedness=False, invert=False):

        "Want all intervals in self that overlap with other."

        df = _overlap(self, other, strandedness, invert)
        return PyRanges(df)


    def nearest(self, other, strandedness=False, is_sorted=False, suffix="_b", how=None):

        "Find the nearest feature in other."

        df = _nearest(self, other, strandedness, suffix, is_sorted, how)

        return PyRanges(df)


    def intersection(self, other, strandedness=False):

        "Want the parts of the intervals in self that overlap with other."

        df = _intersection(self, other, strandedness)

        df.index.name = ""
        return PyRanges(df)


    def set_intersection(self, other, strandedness=False):

        si = _set_intersection(self, other, strandedness)

        return PyRanges(si)


    def set_union(self, other, strand=False):

        si = _set_union(self, other, strand)

        return PyRanges(si)


    def subtraction(self, other, strandedness=False):

        return PyRanges(_subtraction(self, other, strandedness))


    def join(self, other, strandedness=False, new_pos=None, suffixes=["_a", "_b"]):

        df = _overlap_write_both(self, other, strandedness, new_pos, suffixes)

        return PyRanges(df)


    def cluster(self, strand=None, df_only=False, max_dist=0, min_nb=1):

        df = _cluster(self, strand, max_dist, min_nb)

        if not df_only:
            return PyRanges(df)
        else:
            return df


    def tile(self, tile_size=50):

        df = _tile(self, tile_size)
        return PyRanges(df)


    def coverage(self, value_col=None, stranded=False):

        return _coverage(self, value_col, stranded=stranded)



import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture
def expected_result_unstranded():

 c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 6 7 f 0 - 0
1 chr1 5 7 h 0 - 6 7 f 0 - 0"""

 return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_unstranded(f1, f2, expected_result_unstranded):

    result = f1.nearest(f2, strandedness=False, suffix="_b", how="next")

    print(result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_unstranded.df)


@pytest.fixture
def expected_result_opposite_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 6 7 f 0 - 0"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_opposite_stranded(f1, f2, expected_result_opposite_stranded):

    result = f1.nearest(f2, strandedness="opposite", suffix="_b", how="next")

    print(result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_opposite_stranded.df)


@pytest.fixture
def expected_result_same_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 5 7 h 0 - 6 7 f 0 - 0"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_same_stranded(f1, f2, expected_result_same_stranded):

    result = f1.nearest(f2, strandedness="same", suffix="_b", how="next")

    print(result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_same_stranded.df)



import pytest

import pandas as pd

from io import StringIO

from pyranges import PyRanges

from tests.helpers import assert_df_equal

@pytest.fixture
def expected_result_previous_bed_opposite_stranded(names):

    c = """chr1 3 6 h 0 + 6 7 f 0 - 0
chr1 8 9 h 0 + 6 7 f 0 - 1
chr1 5 7 h 0 - 1 2 f 0 + 3"""

    df = pd.read_table(StringIO(c), sep=" ", header=None, names="Chromosome  Start  End  Name Score Strand Start_b  End_b  Name_b Score_b Strand_b Distance".split())
    print(df)

    return PyRanges(df)

def test_nearest_previous_bed_opposite_stranded(f1, f2, expected_result_previous_bed_opposite_stranded):

    print(f1)
    print(f2)
    print(expected_result_previous_bed_opposite_stranded)
    result = f1.nearest(f2, strandedness="opposite", suffix="_b", how="previous")

    print(result.df.to_csv(sep=" "))

    assert assert_df_equal(result.df, expected_result_previous_bed_opposite_stranded.df)


@pytest.fixture
def expected_result_previous_bed_same_stranded(names):

    c = """0 chr1 3 6 h 0 + 1 2 f 0 + 1
1 chr1 5 7 h 0 - 6 7 f 0 - 0
2 chr1 8 9 h 0 + 1 2 f 0 + 6"""

    df = pd.read_table(StringIO(c), sep=" ", header=None, names="Chromosome  Start  End  Name Score Strand Start_b  End_b  Name_b Score_b Strand_b Distance".split())
    print(df)

    return PyRanges(df)


def test_nearest_previous_bed_same_stranded(f1, f2, expected_result_previous_bed_same_stranded):

    result = f1.nearest(f2, strandedness="same", suffix="_b", how="previous")

    assert assert_df_equal(result.df, expected_result_previous_bed_same_stranded.df)


@pytest.fixture
def expected_result_previous_bed_unstranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 6 7 f 0 - 0
1 chr1 5 7 h 0 - 6 7 f 0 - 0
2 chr1 8 9 h 0 + 6 7 f 0 - 1"""

    df = pd.read_table(StringIO(c), sep=" ", header=0)
    print(df)

    return PyRanges(df)

def test_nearest_previous_bed_unstranded(f1, f2, expected_result_previous_bed_unstranded):

    result = f1.nearest(f2, strandedness=None, suffix="_b", how="previous")

    assert assert_df_equal(result.df, expected_result_previous_bed_unstranded.df)



import pandas as pd


def assert_df_equal(df1, df2):

    if "Strand" in df1 and "Strand" in df2:
        sort_on = "Chromosome Start End Strand".split()
    else:
        sort_on = "Chromosome Start End".split()

    df1 = df1.sort_values(sort_on)
    df2 = df2.sort_values(sort_on)

    df1 = df1.reset_index(drop=True)
    df2 = df2.reset_index(drop=True)

    print("Actual")
    print(df1.to_csv(sep=" "))
    print("Expected")
    print(df2.to_csv(sep=" "))

    print("Actual dtypes")
    print(df1.dtypes)
    print("Expected dtypes")
    print(df2.dtypes)
    print("dtypes equal", df1.dtypes == df2.dtypes)

    print("Actual index")
    print(df1.index)
    print("Expected index")
    print(df2.index)
    print("index equal", df1.index == df2.index)

    return df1.equals(df2)
